"""
This client matches https://github.com/afermg/trackastra/blob/main/server.py.

This functions define how Nahual's trackastra client contacts its server counterpart.
"""

import json

import numpy

from nahual.serial import serialize_numpy
from nahual.transport import request_receive


def load_model(
    parameters: dict = {"model": "general_2d", "mode": "greedy"}, address: str = None
) -> str:
    """
    Load a model by sending parameters to a specified address.

    Parameters:
    -----------
    parameters : dict, optional
        A dictionary containing model parameters. Defaults to {"model_name": "general_2d"}.
    address : str
        The address to send the request.

    Returns:
    --------
    str
        The decoded response from the server.
    """
    # encode
    packet = json.dumps(parameters).encode()
    # Request->receive
    response = request_receive(packet, address=address)
    # decode
    decoded = response.decode()
    print(f"REQ: RECEIVED {decoded}")

    return json.loads(decoded)


def process_data(data: list | numpy.ndarray, address: str = None) -> dict:
    """
    Process data by sending it to a specified address.

    Parameters:
    -----------
    data : list or numpy.ndarray
        The input data to process.
    address : str
        The address to send the request.

    Returns:
    --------
    dict
        The processed data received from the server.
    """
    # format
    original_array = numpy.asarray(data)
    # encode
    packet = serialize_numpy(original_array)
    # Request->receive
    response = request_receive(packet, address=address)
    # decode
    decoded = json.loads(response.decode())
    print(f"REQ: RECEIVED PROCESSED DATA {len(decoded), decoded.keys()}")

    return decoded



"""
This client matches https://github.com/afermg/trackastra/blob/main/server.py.

This functions define how Nahual's trackastra client contacts its server counterpart.
"""

import json
import sys

import numpy

from nahual.serial import deserialize_numpy, serialize_numpy
from nahual.transport import request_receive


def load_model(
    parameters: dict = {
        "repo_or_dir": "facebookresearch/dinov2",
        "model": "dinov2_vits14_lc",
    },
    address: str = None,
) -> str:
    """
    Load a model by sending parameters to a specified address.

    Parameters:
    -----------
    parameters : dict, optional
        A dictionary containing model parameters.
    address : str, optional
        The address to send the request. Defaults to ADDRESS.

    Returns:
    --------
    str
        The decoded response from the server.
    """
    # encode
    packet = json.dumps(parameters).encode()
    # Request->receive
    response = request_receive(packet, address=address)
    # decode
    decoded = response.decode()
    print(f"REQ: RECEIVED {decoded}")

    return json.loads(decoded)


def process_data(data: list | numpy.ndarray, address: str = None) -> dict:
    """
    Process data by sending it to a specified address.

    Parameters:
    -----------
    data : list or numpy.ndarray
        The input data to process.
    address : str, optional
        The address to send the request. Defaults to ADDRESS.

    Returns:
    --------
    dict
        The processed data received from the server.
    """
    # format
    original_array = numpy.asarray(data)
    # encode
    packet = serialize_numpy(original_array)
    # Request->receive
    response = request_receive(packet, address=address)
    # decode
    decoded = deserialize_numpy(response)
    print(f"REQ: RECEIVED PROCESSED DATA {decoded.shape, decoded.dtype}")

    return decoded



import pynng


def request_receive(packet: bytes, address: str) -> bytes:
    """Send a request and receive a response using pynng.

    Parameters
    ----------
    packet : bytes
        The data to send in the request.
    address : str, optional
        The endpoint address to connect to, such as "ipc:///tmp/reqrep.ipc".

    Returns
    -------
    bytes
        The response data received from the server.

    Raises
    ------
    pynng.exceptions.Timeout
        If the request times out.

    Notes
    ------
    While the server responder may be async, this method is desygned to run
    in one thread.

    """
    with pynng.Req0() as sock:
        sock.dial(address)
        print(f"REQ: SENDING {len(packet)} bytes")
        sock.send(packet)
        response = sock.recv_msg()
        response_bytes = response.bytes
        print(f"REQ: RECEIVED {len(response_bytes)} bytes")
        return response_bytes



# If you publish results that make use of this software or the Birth Annotator
# for Budding Yeast algorithm, please cite:
# Pietsch, J.M.J., Muñoz, A.F., Adjavon, D.-Y.A., Farquhar, I., Clark, I.B.N.,
# and Swain, P.S. (2023). Determining growth rates from bright-field images of
# budding cells through identifying overlaps. eLife. 12:e79812.
# https://doi.org/10.7554/eLife.79812
#
#
# The MIT License (MIT)
#
# Copyright (c) Julian Pietsch, Alán Muñoz and Diane Adjavon 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
import numpy as np
from matplotlib.colors import Normalize
from matplotlib import pyplot as plt


def colour_seg(seg, rgb=(0.5, 0, 0.5), alpha=1):
    seg = (seg > 0).astype("float")
    img = np.zeros(seg.shape + (4,))
    for i, c in enumerate(rgb):
        img[:, :, i] = c * seg
    img[:, :, -1] = alpha * seg
    return img


def colour_segstack(
    seg_stack, cmap=plt.cm.jet, alpha=0.7, labels=None, max_label=None, dw=False
):
    """Convert a stack of logical masks to coloured images for display

    :param seg_stack: segmentation masks specified index-wise, i.e., in an
        ndarray with shape (N_masks, X, Y)
    :param cmap: one of the matplotlib colour maps
    :param alpha: alpha transparency value between 0 and 1
    :param labels: optionally specify a list of integer cell labels here, one
        for each mask, to ensure consistent colouring of masks by track ID
    :param max_label: optionally specify the maximum expected tracking label
    :param dw: set to True if `seg_stack` is specified depth-wise, i.e., in an
        ndarray with shape (X, Y, N_masks)
    """
    if not dw:
        seg_stack = seg_stack.transpose([1, 2, 0])

    if labels is not None:
        if max_label is None:
            max_label = max(labels)
        seg_stack_orig = seg_stack
        seg_stack = np.zeros(
            seg_stack_orig.shape[:2] + (max_label,), dtype=seg_stack_orig.dtype
        )
        seg_stack[..., np.array(labels) - 1] = seg_stack_orig

    if max_label is None:
        max_label = seg_stack.shape[2]

    seg_enum = (seg_stack > 0) * (1 + np.arange(max_label, dtype=np.uint16))
    seg_flat = np.apply_along_axis(np.max, 2, seg_enum)
    seg_show = Normalize(vmin=1, vmax=max_label)(seg_flat)
    seg_show = cmap(seg_show)
    seg_show[:, :, -1] = alpha * (seg_flat > 0)
    return seg_show


def plot_ims(ims, size=4, cmap=plt.cm.gray, show=True, dw=False, **kwargs):
    if dw:
        ims = ims.transpose([2, 0, 1])

    aspect = ims.shape[2] / ims.shape[1]
    ncols = len(ims)
    fig, axs = plt.subplots(
        1, ncols, figsize=(aspect * ncols * size, size), squeeze=False
    )
    fig.subplots_adjust(wspace=0.02)
    axs = axs[0]
    for ax, im in zip(axs, ims):
        ax.imshow(np.squeeze(im), cmap=cmap, **kwargs)
        ax.grid(False)
        ax.set_xticks([])
        ax.set_yticks([])
    if show:
        plt.show()
    else:
        return fig, axs


def plot_IoU(target_seg, ref_segs, index):
    if index < 0:
        match = np.zeros_like(target_seg)
        others = ref_segs + [np.zeros_like(target_seg)]
    else:
        match = ref_segs[index]
        others = ref_segs[:index] + ref_segs[index + 1 :]

    others = np.max(np.stack(others), axis=0)

    plt.imshow(colour_seg(others, rgb=(0, 0.3, 0), alpha=0.3))
    plt.imshow(colour_seg(match, rgb=(0, 0, 1), alpha=0.5))
    plt.imshow(colour_seg(target_seg, rgb=(1, 0, 0), alpha=0.5))
    plt.grid(False)
    plt.xticks([])
    plt.yticks([])


def plot_overlaps(model, inputim, target, output):
    lblind = model.output_names.index(output)
    target = target[lblind][..., 0] > 0
    pred = model.predict(inputim)[lblind]
    pred_bin = np.squeeze(pred) > 0.5
    correct = pred_bin & target
    falsepos = pred_bin & ~target
    falseneg = ~pred_bin & target

    ncols = len(inputim)
    plt.figure(figsize=(ncols * 4, 4))
    for i in range(ncols):
        plt.subplot(1, ncols, i + 1)
        plt.imshow(np.squeeze(inputim[i, :, :, 0]), cmap="gray")
        plt.imshow(colour_seg(correct[i], rgb=(0, 1, 0), alpha=0.3))
        plt.imshow(colour_seg(falsepos[i], rgb=(0, 0, 1), alpha=0.3))
        plt.imshow(colour_seg(falseneg[i], rgb=(1, 0, 0), alpha=0.3))
        plt.grid(False)
        plt.xticks([])
        plt.yticks([])
    plt.show()


def plot_best_or_worst(outlist):
    ncols = len(outlist)
    plt.figure(figsize=(ncols * 4, 4))
    for i in range(ncols):
        iou, batch_ind, pred, target, inputim = outlist[i]

        correct = pred & target
        falsepos = pred & ~target
        falseneg = ~pred & target

        plt.subplot(1, ncols, i + 1)
        plt.imshow(np.squeeze(inputim[:, :, 0]), cmap="gray")
        plt.imshow(colour_seg(correct, rgb=(0, 1, 0), alpha=0.3))
        plt.imshow(colour_seg(falsepos, rgb=(0, 0, 1), alpha=0.3))
        plt.imshow(colour_seg(falseneg, rgb=(1, 0, 0), alpha=0.3))
        plt.grid(False)
        plt.xticks([])
        plt.yticks([])
        plt.title("IoU: {:.2f}; Batch: {}".format(iou, batch_ind))
    plt.show()



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.2
#   kernelspec:
#     display_name: baby
#     language: python
#     name: baby
# ---

# %%
import requests

baby_url = "http://0.0.0.0:5101"  # URL to reach baby-phone

# %% [markdown]
# ## Check version of the BABY server

# %%
r = requests.get(baby_url)
r.json() if r.ok else r.text

# %% [markdown]
# ## Check which models are available

# %% [markdown]
# ### Summary

# %%
r = requests.get(baby_url + "/models")
r.json() if r.ok else r.text

# %% [markdown]
# ### Including additional meta data

# %%
r = requests.get(baby_url + "/models?meta=true")
r.json() if r.ok else r.text

# %% [markdown]
# ## Start a new session

# %%
modelset = "yeast-alcatras-brightfield-sCMOS-60x-1z"
r = requests.get(f"{baby_url}/session/{modelset}")
if not r.ok:
    raise Exception(f"{r.status_code}: {r.text}")
print(r.json())
session_id = r.json()["sessionid"]

# %% [markdown]
# ## List running sessions

# %%
r = requests.get(baby_url + "/sessions")
r.json() if r.ok else r.text

# %% [markdown]
# ## Send an image to the server

# %%
import numpy as np

rng = np.random.default_rng(42)

# Create suitable N x H x W x Z array
# dtype must be either uint8 or uint16
img = rng.integers(2**8, size=(2, 80, 120, 1), dtype="uint8")

# Initiate a multipart-encoded request
r = requests.post(
    f"{baby_url}/segment?sessionid={session_id}",
    files=[
        # The ordering of these parts must be fixed
        ("dims", ("", str(list(img.shape)))),
        ("bitdepth", ("", "8")),
        ("img", ("", img.tobytes(order="F"))),
        # Optionally specify additional parts that set
        # any kwargs accepted by BabyCrawler.step (ordering
        # is no longer fixed)
        ("refine_outlines", ("", "true")),
    ],
)

r.json() if r.ok else r.text

# %% [markdown]
# *NB: To run the following, `baby-phone` should have been started in debug mode (`baby-phone --debug-mode`) in a subfolder named `baby-phone-debugging`.*

# %%
from imageio import v3 as iio

img1 = iio.imread("../../../baby_train/baby_phone_debugging_image_received_001.png")
# Check that the image received by the server is identical to the one sent
assert np.allclose(img1, img[0, ..., -1])
# metadata associated with the image (i.e., info1) includes the additional keyword arguments
img1.shape, img1.dtype

# %% [markdown]
# ## Retrieve results after processing

# %% [markdown]
# ### Default output without edge masks

# %% [markdown]
# By default the returned result does not include the edge masks and the outlines need to be recreated by the client from the radial spline representation (e.g., using `baby.segmentation.draw_radial`). This is how I did it for the Matlab client to reduce the amount of data being sent over the socket.
#
# See the next section for an example including edge mask output.

# %%
r = requests.get(f"{baby_url}/segment?sessionid={session_id}")
if not r.ok:
    raise Exception(f"{r.status_code}: {r.text}")
outputs = r.json()

# outputs is a list of dicts for each image in batch
print([{k: len(v) for k, v in output.items()} for output in outputs])

# The outputs are (lists of) lists of numbers
with np.printoptions(precision=2):
    print("\ncentres:")
    print(np.array(outputs[0]["centres"]))  # row, col
    print("\nradii:")
    print([np.array(r) for r in outputs[0]["radii"]])

# %% [markdown]
# ### Process images to get output *with* edge masks

# %% [markdown]
# Process the images with the keyword argument `with_edgemasks` set to true:

# %%
# Initiate a multipart-encoded request
r = requests.post(
    f"{baby_url}/segment?sessionid={session_id}",
    files=[
        # The ordering of these parts must be fixed
        ("dims", ("", str(list(img.shape)))),
        ("bitdepth", ("", "8")),
        ("img", ("", img.tobytes(order="F"))),
        # Optionally specify additional parts that set
        # any kwargs accepted by BabyCrawler.step (ordering
        # is no longer fixed)
        ("refine_outlines", ("", "true")),
        ("with_edgemasks", ("", "true")),
    ],
)

r.json() if r.ok else r.text

# %% [markdown]
# Retrieve the processed results:

# %%
r = requests.get(f"{baby_url}/segment?sessionid={session_id}")
if not r.ok:
    raise Exception(f"{r.status_code}: {r.text}")
outputs = r.json()
len(outputs), outputs[0].keys()

# %% [markdown]
# The edgemasks are returned as coordinates using Matlab indexing (i.e., starting from 1). We can regenerate edge mask images from the coordinates using the following:

# %%
all_edgemasks = []
for output in outputs:
    edge_coords_i = output["edgemasks"]
    edgemasks_i = []
    for edge_coords in edge_coords_i:
        edgemask = np.zeros(img.shape[1:3], dtype="bool")
        edgemask[tuple(np.array(d) - 1 for d in edge_coords)] = True
        edgemasks_i.append(edgemask)
    all_edgemasks.append(np.stack(edgemasks_i))

# %% [markdown]
# Now we can display the outlines that BABY hallucinated from noise:

# %%
from matplotlib import pyplot as plt
from visualise import colour_segstack, plot_ims

_, axs = plot_ims(img.squeeze(), show=False)
for ax, edgemasks in zip(axs, all_edgemasks):
    ax.imshow(colour_segstack(edgemasks))
plt.show()



import struct
from itertools import chain

import numpy


def serialize_numpy(data: numpy.ndarray) -> bytes:
    """Serialize numpy-like arrays while encoding their dtype and shape.

    Parameters
    ----------
    data : numpy.ndarray
        The input array to be serialized.

    Returns
    -------
    bytes
        The serialized array, containing dtype, shape, and data.

    Notes
    -----
    The serialized format is as follows:
    - Byte 0: Datatype represented as a single character
    - Byte 1: Number of dimensions, encoded as an unsigned 8-bit integer.
    - Bytes 2-n: The shape of every dimension, encoded as an unsigned short.
    - Remaining bytes: The contents of the numpy array.
    """
    array_bytes = data.tobytes()
    dtype_char = data.dtype.char.encode()
    ndim = data.ndim.to_bytes()
    shape = data.shape

    shape = bytes()
    for x in data.shape:
        shape = shape + struct.pack(">H", x)
    message = dtype_char + ndim + shape + array_bytes

    return message


def deserialize_numpy(message: bytes) -> numpy.ndarray:
    """Deserialize a NumPy array from a byte stream.

    Parameters
    ----------
    message : bytes
        The byte stream containing the serialized NumPy array.

    Returns
    -------
    numpy.ndarray
        The deserialized NumPy array.

    Notes
    -----
    The byte stream format is as follows:
    - First byte: data type (dtype)
    - Second byte: number of dimensions (ndim)
    - Remaining bytes: shape of the array
    """
    # First byte: dtype char
    in_dtype = numpy.dtype(message[:1].decode())
    # Second byte: ndim
    in_ndim = int.from_bytes(message[1:2])
    # Third->n: shape
    in_shape = tuple(
        chain.from_iterable([
            struct.unpack(">H", message[i * 2 + 2 : i * 2 + 4]) for i in range(in_ndim)
        ])
    )

    reshaped_array = numpy.frombuffer(
        message[2 * in_ndim + 2 :], dtype=in_dtype
    ).reshape(in_shape)

    return reshaped_array
